
import logging
import json

def mask_bearer_token(token: str) -> str:
    if not token.startswith("Bearer "):
        return token
    raw = token[7:]  # remove "Bearer "
    if len(raw) <= 8:
        return "Bearer ****"
    return f"Bearer {raw[:4]}****{raw[-4:]}"


def logRequestPayload(debug, payload, url, headers):
    logging.debug("POST %s", url)
    if debug:
        # Mask sensitive header before logging
        masked_headers = {
            **headers,
            "Authorization": mask_bearer_token(headers["Authorization"])
        }
        logging.debug("Request Headers:\n%s", json.dumps(masked_headers, indent=2))
    logging.debug("Request Payload:\n%s", json.dumps(payload, indent=2))     


def configure_logging(debug: bool):
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.ERROR,
        format="%(asctime)s - %(levelname)s - %(message)s"
    )


"""PayPal Agentic Tools."""
import json
from agents import FunctionTool
from agents.run_context import RunContextWrapper

from ..shared.api import PayPalAPI

def PayPalTool(api: PayPalAPI, tool) -> FunctionTool:
    async def on_invoke_tool(ctx: RunContextWrapper, input_str: str) -> str:
        return api.run(tool["method"], json.loads(input_str))

    parameters = tool["args_schema"].model_json_schema()
    
    # Enforce schema constraints
    parameters.update({
        "additionalProperties": False,
        "type": "object"
    })

    # Remove unnecessary metadata
    for key in ["description", "title"]:
        parameters.pop(key, None)

    # Clean up properties if they exist
    for prop in parameters.get("properties", {}).values():
        for key in ["title", "default"]:
            prop.pop(key, None)

    return FunctionTool(
        name=tool["method"],
        description=tool["description"],
        params_json_schema=parameters,
        on_invoke_tool=on_invoke_tool,
        strict_json_schema=False
    )


import json
from typing import Optional
import requests

from ..shared.telemetry import Telemetry

from .logger_util import configure_logging, logRequestPayload
from .constants import *
from .configuration import Context
import logging


class PayPalClient:
    def __init__(self, client_id, secret, context: Optional[Context]):
        self.client_id = client_id
        self.secret = secret
        self.context = context
        self.sandbox = context.sandbox
        self.debug = context.debug
        self.base_url = SANDBOX_BASE_URL if self.sandbox  else LIVE_BASE_URL
        configure_logging(self.debug)


    def log_request_exception(self, e: requests.exceptions.RequestException, url: Optional[str] = None):
        response = getattr(e, 'response', None)
        if response is not None:
            try:
                logging.error("PayPal Error Response: %s", json.dumps(response.json(), indent=2))
            except ValueError:
                logging.error("PayPal Error Response: Not valid JSON")
            logging.error("Response Headers: %s", json.dumps(dict(response.headers), indent=2))
        if url:
            logging.error("Request to %s failed: %s", url, str(e))
        else:
            logging.error("HTTP request failed: %s", str(e))

    def build_headers(self):
        headers = {
                "Authorization": f"Bearer {self.get_access_token()}",
                "Content-Type": "application/json",
                "User-Agent" : Telemetry.generate_user_agent(source=self.context.source)
            }
        return headers

    def get_access_token(self):
        token_url = f"{self.base_url}/v1/oauth2/token"
        try:
            response = requests.post(
                token_url,
                headers={"Accept": "application/json"},
                data={"grant_type": "client_credentials"},
                auth=(self.client_id, self.secret)
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            self.log_request_exception(e, token_url)
            raise RuntimeError("Failed to obtain access token from PayPal") from e
        
        if self.debug:
            logging.debug("PayPal Response Headers: %s", json.dumps(dict(response.headers), indent=2))

        token_data = response.json()
        if "access_token" not in token_data:
            raise ValueError("Access token not found in PayPal response")

        return token_data["access_token"]


    def post(self, uri, payload):
       
        url = f"{self.base_url}{uri}"
        headers = self.build_headers()
        logRequestPayload(self.debug, payload, url, headers)

        try:
            response = requests.post(url, headers=headers, json=payload)
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            self.log_request_exception(e, url)
            raise

        if response.status_code == 204:
            logging.debug("Response Status: 204 No Content")
            return {}
        
        try:
            json_response = response.json()
        except ValueError:
            logging.warning("Response body is not valid JSON or empty, Headers: %s", json.dumps(dict(response.headers), indent=2))
            return {}

        if self.debug:
            logging.debug("PayPal Response Headers: %s", json.dumps(dict(response.headers), indent=2))
            logging.debug("PayPal Response Payload: %s", json.dumps(json_response, indent=2))

        return json_response
    

    def get(self, uri):

        url = f"{self.base_url}{uri}"
        headers = self.build_headers()
        
        logRequestPayload(self.debug, None, url, headers)

        try:
            response = requests.get(url, headers=headers)
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            self.log_request_exception(e, url)
            logging.error("HTTP request failed: %s", str(e))
            raise

        try:
            json_response = response.json()
        except ValueError:
            logging.warning("Response body is not valid JSON or empty")
            return {}

        if self.debug:
            logging.debug("PayPal Response Headers: %s", json.dumps(dict(response.headers), indent=2))
            logging.debug("PayPal Response Payload: %s", json.dumps(json_response, indent=2))

        return json_response



    
    





from typing import Optional
from pydantic import BaseModel
from .configuration import Context
from .paypal_client import PayPalClient
from .tools import tools

class PayPalAPI(BaseModel):
    
    _context: Context
    _paypal_client: PayPalClient
    
    def __init__(self, client_id: str, secret: str, context: Optional[Context]):
        super().__init__()

        self._context = context if context is not None else Context()
        self._paypal_client = PayPalClient(client_id=client_id, secret=secret, context=context)
        
    
    def run(self, method: str, params: dict) -> str:
        for tool in tools:
            if tool.get("method") == method:
                execute_fn = tool.get("execute")
                if execute_fn:
                    return execute_fn(self._paypal_client, params)
        raise ValueError(f"method: {method} not found in tools list")




from pydantic import BaseModel, Field, HttpUrl, validator, field_validator, ConfigDict, constr
from typing import List, Literal, Optional



class ItemDetails(BaseModel):
    item_cost: float = Field(..., description="The cost of each item – up to 2 decimal points.")
    tax_percent: float = Field(0, description="The tax percent for the specific item.")
    item_total: float = Field(..., description="The total cost of this line item.")


class LineItem(ItemDetails):
    name: str = Field(..., description="The name of the item.")
    quantity: int = Field(
        1,
        description="The item quantity. Must be a whole number.",
        ge=1
    )
    description: Optional[str] = Field(
        None,
        description="The detailed item description."
    )


class ShippingAddress(BaseModel):
    address_line_1: Optional[str] = Field(
        None,
        description=(
            "The first line of the address, such as number and street, "
            "for example, `173 Drury Lane`. This field needs to pass the full address."
        )
    )
    address_line_2: Optional[str] = Field(
        None,
        description="The second line of the address, for example, a suite or apartment number."
    )
    admin_area_2: Optional[str] = Field(
        None,
        description="A city, town, or village. Smaller than `admin_area_level_1`."
    )
    admin_area_1: Optional[str] = Field(
        None,
        description=(
            "The highest-level sub-division in a country, which is usually a province, "
            "state, or ISO-3166-2 subdivision."
        )
    )
    postal_code: Optional[str] = Field(
        None,
        description=(
            "The postal code, which is the ZIP code or equivalent. Typically required "
            "for countries with a postal code or an equivalent."
        )
    )
    country_code: Optional[constr(min_length=2, max_length=2)] = Field(
        None,
        description=(
            "The 2-character ISO 3166-1 code that identifies the country or region. "
            "Note: The country code for Great Britain is `GB`."
        )
    )


class CreateOrderParameters(BaseModel):
    model_config = ConfigDict(validate_default=True)
    currency_code: Literal["USD"] = Field(
        ...,
        description="Currency code of the amount."
    )
    items: List[LineItem] = Field(
        ...,
        description="List of individual items in the order (max 50)."
    )
    discount: float = Field(
        0,
        description="The discount amount for the order."
    )
    shipping_cost: float = Field(
        0,
        description="The cost of shipping for the order."
    )
    shipping_address: Optional[ShippingAddress] = Field(
        None,
        description="The shipping address for the order."
    )
    notes: Optional[str] = Field(
        None,
        description="Optional customer notes or instructions."
    )
    return_url: Optional[HttpUrl] = Field(
        "https://example.com/returnUrl",
        description="URL to redirect the buyer after approval."
    )
    cancel_url: Optional[HttpUrl] = Field(
        "https://example.com/cancelUrl",
        description="URL to redirect the buyer if they cancel."
    )


class OrderIdParameters(BaseModel):
    order_id: str

class CaptureOrderParameters(BaseModel):
    order_id: str
    


import logging
import re
from urllib.parse import urlencode

def parse_order_details(params: dict) -> dict:
    try:
        # use snake_case keys
        curr_code     = params["currency_code"]
        items_in      = params.get("items", [])
        shipping_cost = params.get("shipping_cost", 0)
        discount      = params.get("discount", 0)

        sub_total = sum(item["item_cost"] * item.get("quantity", 1) for item in items_in)
        tax_amount = sum(
            item["item_cost"] * item.get("tax_percent", 0) * item.get("quantity", 1) / 100
            for item in items_in
        )
        total = sub_total + tax_amount + shipping_cost - discount

        amount_breakdown = {
            "item_total": {
                "value": f"{sub_total:.2f}",
                "currency_code": curr_code,
            },
            "shipping": {
                "value": f"{shipping_cost:.2f}",
                "currency_code": curr_code,
            },
            "tax_total": {
                "value": f"{tax_amount:.2f}",
                "currency_code": curr_code,
            },
            "discount": {
                "value": f"{discount:.2f}",
                "currency_code": curr_code,
            },
        }

        items = []
        for item in items_in:
            items.append({
                "name": item["name"],
                "description": item.get("description", ""),
                "unit_amount": {
                    "value": f"{item['item_cost']:.2f}",
                    "currency_code": curr_code,
                },
                "quantity": str(item.get("quantity", 1)),
                "tax": {
                    "value": f"{(item['item_cost'] * item.get('tax_percent', 0) / 100):.2f}",
                    "currency_code": curr_code,
                }
            })

        base_purchase_unit = {
            "amount": {
                "value": f"{total:.2f}",
                "currency_code": curr_code,
                "breakdown": amount_breakdown,
            },
            "items": items,
        }

        if params.get("shipping_address"):
            base_purchase_unit["shipping"] = {"address": params["shipping_address"]}

        request = {
            "intent": "CAPTURE",
            "purchase_units": [base_purchase_unit],
        }

        experience_context = {}
        if params.get("return_url"):
            experience_context["return_url"] = str(params["return_url"])
        if params.get("cancel_url"):
            experience_context["cancel_url"] = str(params["cancel_url"])

        if experience_context:
            request["payment_source"] = {
                "paypal": {
                    "experience_context": experience_context
                }
            }

        return request

    except Exception as e:
        logging.error("parse_order_details error:", e)
        raise ValueError("Failed to parse order details") from e



def to_snake_case_keys(obj):
    """
    Recursively convert dict keys from camelCase (or mixed) to snake_case.
    """
    if isinstance(obj, list):
        return [to_snake_case_keys(v) for v in obj]
    elif isinstance(obj, dict):
        new = {}
        for k, v in obj.items():
            # insert underscore before capital letters, then lower()
            new_key = re.sub(r"(?<!^)(?=[A-Z])", "_", k).lower()
            new[new_key] = to_snake_case_keys(v)
        return new
    else:
        return obj


def to_camel_case_keys(obj):
    """
    Recursively convert dict keys from snake_case (or mixed) to camelCase.
    """
    if isinstance(obj, list):
        return [to_camel_case_keys(v) for v in obj]
    elif isinstance(obj, dict):
        new = {}
        for k, v in obj.items():
            parts = k.split("_")
            camel = parts[0] + "".join(word.title() for word in parts[1:])
            new[camel] = to_camel_case_keys(v)
        return new
    else:
        return obj


def to_query_string(params: dict) -> str:
    """
    Build a URL query string from a flat dict, ignoring None values.
    """
    filtered = {k: v for k, v in params.items() if v is not None}
    # urlencode will properly escape keys and values
    return urlencode({k: str(v) for k, v in filtered.items()})




from .parameters import *
import json
import httpx
from typing import Union, Dict, Any



def create_invoice(client, params: dict):
    
    validated = CreateInvoiceParameters(**params)
    invoice_payload = validated.model_dump()

    url = "/v2/invoicing/invoices"
    response = client.post(uri=url, payload=invoice_payload)

    if (
        response.get("rel") == "self"
        and "/v2/invoicing/invoices/" in response.get("href", "")
        and response.get("method") == "GET"
    ):
        invoice_id = response["href"].split("/")[-1]
        try:
            send_result = send_invoice(client, {
                "invoice_id": invoice_id,
                "note": "Thank you for choosing us. If there are any issues, feel free to contact us.",
                "send_to_recipient": True
            })
            return json.dumps({
                "createResult": response,
                "sendResult": send_result
            })
        except Exception:
            return json.dumps(response)

    return json.dumps(response)


def send_invoice(client, params: dict):

    validated = SendInvoiceParameters(**params)
    payload = validated.model_dump()

    invoice_id = payload["invoice_id"]
    url = f"/v2/invoicing/invoices/{invoice_id}/send"

    response =  client.post(uri=url, payload=payload)
    return json.dumps(response)


def list_invoices(client, params: dict):

    validated = ListInvoicesParameters(**params)
    invoice_uri = f"/v2/invoicing/invoices?page_size={validated.page_size or 10}&page={validated.page or 1}&total_required={validated.total_required or 'true'}"
    response = client.get(uri=invoice_uri)

    return json.dumps(response)


def get_invoice(client, params: dict):
    validated = GetInvoiceParameters(**params)
    invoice_id = validated.invoice_id

    url = f"/v2/invoicing/invoices/{invoice_id}"
    response = client.get(uri=url)

    return json.dumps(response)


def send_invoice_reminder(client, params: dict):

    validated = SendInvoiceReminderParameters(**params)
    payload = validated.model_dump()

    invoice_id = payload["invoice_id"]
    url = f"/v2/invoicing/invoices/{invoice_id}/remind"
    print("url: ", url)
    response = client.post(uri=url, payload=payload)
    print("response: ", response)

    if response is None:
        return {"success": True, "invoice_id": invoice_id}
    return json.dumps(response)


def cancel_sent_invoice(client, params: dict):
    
    validated = CancelSentInvoiceParameters(**params)
    payload = validated.model_dump()
    invoice_id = payload["invoice_id"]
    url = f"/v2/invoicing/invoices/{invoice_id}/cancel"

    response = client.post(uri=url, payload=payload)
    
    # PayPal responds with 204 No Content on successful cancellation
    if response is None:
        return {"success": True, "invoice_id": invoice_id}

    return json.dumps(response)


def generate_invoice_qrcode(client, params: dict):

    validated = GenerateInvoiceQrCodeParameters(**params)
    payload = {
        "width": validated.width,
        "height": validated.height
    }

    invoice_id = validated.invoice_id
    url = f"/v2/invoicing/invoices/{invoice_id}/generate-qr-code"

    response = client.post(uri=url, payload=payload)

    if response is None:
        return {"success": True, "invoice_id": invoice_id}

    return json.dumps(response)



from .parameters import *
import json

 
def create_product(client, params: dict):

    validated = CreateProductParameters(**params)
    product_uri = "/v1/catalogs/products"
    result = client.post(uri = product_uri, payload = validated.model_dump())
    return json.dumps(result)


def list_products(client, params: dict):

    validated = ListProductsParameters(**params)
    product_uri = f"/v1/catalogs/products?page_size={validated.page_size or 10}&page={validated.page or 1}&total_required={validated.total_required or 'true'}"
    result = client.get(uri = product_uri)
    return json.dumps(result)


def show_product_details(client, params: dict):

    validated = ShowProductDetailsParameters(**params)
    product_uri = f"/v1/catalogs/products/{validated.product_id}"
    result = client.get(uri = product_uri)
    return json.dumps(result)


def create_subscription_plan(client, params: dict):

    validated = CreateSubscriptionPlanParameters(**params)
    subscription_plan_uri = "/v1/billing/plans"
    result = client.post(uri = subscription_plan_uri, payload = validated.model_dump())
    return json.dumps(result)


def list_subscription_plans(client, params: dict):

    validated = ListSubscriptionPlansParameters(**params)
    subscription_plan_uri = f"/v1/billing/plans?page_size={validated.page_size or 10}&page={validated.page or 1}&total_required={validated.total_required or True}"
    if validated.product_id:
        subscription_plan_uri += f"&product_id={validated.product_id}"
    result = client.get(uri = subscription_plan_uri)
    return json.dumps(result)


def show_subscription_plan_details(client, params: dict):

    validated = ShowSubscriptionPlanDetailsParameters(**params)
    subscription_plan_uri = f"/v1/billing/plans/{validated.plan_id}"
    result = client.get(uri = subscription_plan_uri)
    return json.dumps(result)


def create_subscription(client, params: dict):

    validated = CreateSubscriptionParameters(**params)
    subscription_plan_uri = "/v1/billing/subscriptions"
    result = client.post(uri = subscription_plan_uri, payload = validated.model_dump())
    return json.dumps(result)


def show_subscription_details(client, params: dict):

    validated = ShowSubscriptionDetailsParameters(**params)
    subscription_plan_uri = f"/v1/billing/subscriptions/{validated.subscription_id}"
    result = client.get(uri = subscription_plan_uri)
    return json.dumps(result)


def cancel_subscription(client, params: dict):

    validated = CancelSubscriptionParameters(**params)
    subscription_plan_uri = f"/v1/billing/subscriptions/{validated.subscription_id}/cancel"
    result = client.post(uri = subscription_plan_uri, payload = validated.payload.model_dump())
    if not result:
        return "Successfully cancelled the subscription."
    return json.dumps(result)


from paypal_agent_toolkit.shared.orders.prompts import (
    CREATE_ORDER_PROMPT,
    CAPTURE_ORDER_PROMPT,
    GET_ORDER_PROMPT,
)
from paypal_agent_toolkit.shared.subscriptions.prompts import (
    CREATE_PRODUCT_PROMPT,
    LIST_PRODUCTS_PROMPT,
    SHOW_PRODUCT_DETAILS_PROMPT,
    CREATE_SUBSCRIPTION_PLAN_PROMPT,
    LIST_SUBSCRIPTION_PLANS_PROMPT,
    SHOW_SUBSCRIPTION_PLAN_DETAILS_PROMPT,
    CREATE_SUBSCRIPTION_PROMPT,
    SHOW_SUBSCRIPTION_DETAILS_PROMPT,
    CANCEL_SUBSCRIPTION_PROMPT,
)

from paypal_agent_toolkit.shared.invoices.prompts import (
    CREATE_INVOICE_PROMPT,
    LIST_INVOICE_PROMPT,
    GET_INVOICE_PROMPT,
    SEND_INVOICE_PROMPT,
    SEND_INVOICE_REMINDER_PROMPT,
    CANCEL_SENT_INVOICE_PROMPT,
    GENERATE_INVOICE_QRCODE_PROMPT,
)

from paypal_agent_toolkit.shared.orders.parameters import (
    
    CreateOrderParameters,
    OrderIdParameters,
)

from paypal_agent_toolkit.shared.subscriptions.parameters import (
    
    CreateProductParameters,
    ListProductsParameters,
    ShowProductDetailsParameters,
    CreateSubscriptionPlanParameters,
    ListSubscriptionPlansParameters,
    ShowSubscriptionPlanDetailsParameters,
    CreateSubscriptionParameters,
    ShowSubscriptionDetailsParameters,
    CancelSubscriptionParameters,
)

from paypal_agent_toolkit.shared.invoices.parameters import (
    CreateInvoiceParameters,
    SendInvoiceParameters,
    ListInvoicesParameters,
    GetInvoiceParameters,
    SendInvoiceReminderParameters,
    CancelSentInvoiceParameters,
    GenerateInvoiceQrCodeParameters,
)

from paypal_agent_toolkit.shared.orders.tool_handlers import (
    create_order,
    capture_order,
    get_order_details,
)

from paypal_agent_toolkit.shared.subscriptions.tool_handlers import (
    create_product,
    list_products,
    show_product_details,
    create_subscription_plan,
    list_subscription_plans,
    show_subscription_plan_details,
    create_subscription,
    show_subscription_details,
    cancel_subscription,
)

from paypal_agent_toolkit.shared.invoices.tool_handlers import (
    create_invoice,
    send_invoice,
    list_invoices,
    get_invoice,
    send_invoice_reminder,
    cancel_sent_invoice,
    generate_invoice_qrcode
)

from pydantic import BaseModel

tools = [
    {
        "method": "create_order",
        "name": "Create PayPal Order",
        "description": CREATE_ORDER_PROMPT.strip(),
        "args_schema": CreateOrderParameters,
        "actions": {"orders": {"create": True}},
        "execute": create_order,
    },
    {
        "method": "pay_order",
        "name": "Process payment for PayPal Order",
        "description": CAPTURE_ORDER_PROMPT.strip(),
        "args_schema": OrderIdParameters,
        "actions": {"orders": {"capture": True}},
        "execute": capture_order,
    },
    {
        "method": "get_order_details",
        "name": "Get PayPal Order Details",
        "description": GET_ORDER_PROMPT.strip(),
        "args_schema": OrderIdParameters,
        "actions": {"orders": {"get": True}},
        "execute": get_order_details,
    },
    {
        "method": "create_product",
        "name": "Create PayPal Product",
        "description": CREATE_PRODUCT_PROMPT.strip(),
        "args_schema": CreateProductParameters,
        "actions": {"products": {"create": True}},
        "execute": create_product,
    },
    {
        "method": "list_products",
        "name": "List PayPal Products",
        "description": LIST_PRODUCTS_PROMPT.strip(),
        "args_schema": ListProductsParameters,
        "actions": {"products": {"list": True}},
        "execute": list_products,
    },
    {
        "method": "show_product_details",
        "name": "Show PayPal Product Details",
        "description": SHOW_PRODUCT_DETAILS_PROMPT.strip(),
        "args_schema": ShowProductDetailsParameters,
        "actions": {"products": {"show": True}},
        "execute": show_product_details,
    },
    {
        "method": "create_subscription_plan",
        "name": "Create PayPal Subscription Plan",
        "description": CREATE_SUBSCRIPTION_PLAN_PROMPT.strip(),
        "args_schema": CreateSubscriptionPlanParameters,
        "actions": {"subscriptionPlans": {"create": True}},
        "execute": create_subscription_plan,
    },
    {
        "method": "list_subscription_plans",
        "name": "List PayPal Subscription Plans",
        "description": LIST_SUBSCRIPTION_PLANS_PROMPT.strip(),
        "args_schema": ListSubscriptionPlansParameters,
        "actions": {"subscriptionPlans": {"list": True}},
        "execute": list_subscription_plans,
    },
    {
        "method": "show_subscription_plan_details",
        "name": "List PayPal Subscription Plan Details",
        "description": SHOW_SUBSCRIPTION_PLAN_DETAILS_PROMPT.strip(),
        "args_schema": ShowSubscriptionPlanDetailsParameters,
        "actions": {"subscriptionPlans": {"show": True}},
        "execute": show_subscription_plan_details,
    },
    {
        "method": "create_subscription",
        "name": "Create PayPal Subscription",
        "description": CREATE_SUBSCRIPTION_PROMPT.strip(),
        "args_schema": CreateSubscriptionParameters,
        "actions": {"subscriptions": {"create": True}},
        "execute": create_subscription,
    },
    {
        "method": "show_subscription_details",
        "name": "Show PayPal Subscription Details",
        "description": SHOW_SUBSCRIPTION_DETAILS_PROMPT.strip(),
        "args_schema": ShowSubscriptionDetailsParameters,
        "actions": {"subscriptions": {"show": True}},
        "execute": show_subscription_details,
    },
    {
        "method": "cancel_subscription",
        "name": "Cancel PayPal Subscription",
        "description": CANCEL_SUBSCRIPTION_PROMPT.strip(),
        "args_schema": CancelSubscriptionParameters,
        "actions": {"subscriptions": {"cancel": True}},
        "execute": cancel_subscription,
    },
    {
        "method": "create_invoice",
        "name": "Create PayPal Invoice",
        "description": CREATE_INVOICE_PROMPT.strip(),
        "args_schema": CreateInvoiceParameters,
        "actions": {"invoices": {"create": True}},
        "execute": create_invoice,
    },
    {
        "method": "list_invoices",
        "name": "List Invoices",
        "description": LIST_INVOICE_PROMPT.strip(),
        "args_schema": ListInvoicesParameters,
        "actions": {"invoices": {"list": True}},
        "execute": list_invoices,
    },
    {
        "method": "get_invoice",
        "name": "Get Invoice",
        "description": GET_INVOICE_PROMPT.strip(),
        "args_schema": GetInvoiceParameters,
        "actions": {"invoices": {"get": True}},
        "execute": get_invoice,
    },
    {
        "method": "send_invoice",
        "name": "Send Invoice",
        "description": SEND_INVOICE_PROMPT.strip(),
        "args_schema": SendInvoiceParameters,
        "actions": {"invoices": {"send": True}},
        "execute": send_invoice,
    },
    {
        "method": "send_invoice_reminder",
        "name": "Send Invoice Reminder",
        "description": SEND_INVOICE_REMINDER_PROMPT.strip(),
        "args_schema": SendInvoiceReminderParameters,
        "actions": {"invoices": {"sendReminder": True}},
        "execute": send_invoice_reminder,
    },
    {
        "method": "cancel_sent_invoice",
        "name": "Cancel Sent Invoice",
        "description": CANCEL_SENT_INVOICE_PROMPT.strip(),
        "args_schema": CancelSentInvoiceParameters,
        "actions": {"invoices": {"cancel": True}},
        "execute": cancel_sent_invoice,
    },
    {
        "method": "generate_invoice_qr_code",
        "name": "Generate Invoice QR Code",
        "description": GENERATE_INVOICE_QRCODE_PROMPT.strip(),
        "args_schema": GenerateInvoiceQrCodeParameters,
        "actions": {"invoices": {"generateQRC": True}},
        "execute": generate_invoice_qrcode,
    },
    
]



import logging
import platform
import sys
from importlib.metadata import version, PackageNotFoundError


class Telemetry:
    SDK_NAME = "Paypal Agent Toolkit Python"
    PACKAGE_NAME = "paypal-agent-toolkit"

    @classmethod
    def get_sdk_version(cls) -> str:
        try:
            return version(cls.PACKAGE_NAME)
        except PackageNotFoundError:
            return "unknown"

    @classmethod
    def get_env_info(cls) -> dict:
        return {
            "os": platform.system(),
            "os_version": platform.version(),
            "python_version": sys.version.split()[0],
            "platform": platform.platform(),
            "machine": platform.machine(),
            "hostname": platform.node()
        }


    @classmethod
    def generate_user_agent(cls, source: str) -> str:
        components = [
            f"{cls.SDK_NAME}: {source}",
            f"Version: {cls.get_sdk_version()}",
            f"on OS: {platform.system()} {platform.release()}"
        ]
        return ", ".join(filter(None, components))
    

    @classmethod
    def log(cls):
        telemetry_data = {
            "sdk_name": cls.SDK_NAME,
            "sdk_version": cls.get_sdk_version(),
            **cls.get_env_info()
        }
        logging.debug("Telemetry Info:", telemetry_data) 




from .parameters import *
from .payload_util import parse_order_details
import json

def create_order(client, params: dict):

    validated = CreateOrderParameters(**params)
    order_payload = parse_order_details(validated.model_dump())
    
    order_uri = "/v2/checkout/orders"
    response = client.post(uri=order_uri, payload=order_payload)
    return json.dumps(response)



def capture_order(client, params: dict):
    validated = CaptureOrderParameters(**params)
    order_capture_uri = f"/v2/checkout/orders/{validated.order_id}/capture"
    result = client.post(uri=order_capture_uri, payload=None)
    status = result.get("status")
    amount = result.get("purchase_units", [{}])[0].get("payments", {}).get("captures", [{}])[0].get("amount", {}).get("value")
    currency = result.get("purchase_units", [{}])[0].get("payments", {}).get("captures", [{}])[0].get("amount", {}).get("currency_code")

    return json.dumps({
        "message": f"The PayPal order {validated.order_id} has been successfully captured.",
        "status": status,
        "amount": f"{currency} {amount}" if amount and currency else "N/A",
        "raw": result 
    })


def get_order_details(client, params: dict):
    validated = OrderIdParameters(**params)
    order_get_uri = f"/v2/checkout/orders/{validated.order_id}"
    
    result = client.get(order_get_uri)
    status = result.get("status")
    amount = result.get("purchase_units", [{}])[0].get("payments", {}).get("captures", [{}])[0].get("amount", {}).get("value")
    currency = result.get("purchase_units", [{}])[0].get("payments", {}).get("captures", [{}])[0].get("amount", {}).get("currency_code")

    return json.dumps({
        "message": f"The PayPal order {validated.order_id} has been successfully captured.",
        "status": status,
        "amount": f"{currency} {amount}" if amount and currency else "N/A",
        "raw": result 
    })


from pydantic import BaseModel, Field, HttpUrl, validator, field_validator, ConfigDict, constr
from typing import List, Literal, Optional


class CreateProductParameters(BaseModel):
    name: str
    type: Literal['PHYSICAL', 'DIGITAL', 'SERVICE']  # Enum-like behavior for product type
    description: Optional[str] = None
    category: Optional[str] = None
    image_url: Optional[HttpUrl] = None  # Ensures valid URL
    home_url: Optional[HttpUrl] = None  # Ensures valid URL

class ListProductsParameters(BaseModel):
    page: Optional[int] = None
    page_size: Optional[int] = None
    total_required: Optional[bool] = None

class ShowProductDetailsParameters(BaseModel):
    product_id: str

# Frequency Schema
class FrequencySchema(BaseModel):
    interval_unit: Literal['DAY', 'WEEK', 'MONTH', 'YEAR'] = Field(..., description="The unit of time for the billing cycle.")
    interval_count: int = Field(..., description="The number of units for the billing cycle.")

# Pricing Scheme Schema
class FixedPriceSchema(BaseModel):
    currency_code: Literal['USD'] = Field(..., description="The currency code for the fixed price.")
    value: str = Field(..., description="The value of the fixed price.")

class PricingSchemeSchema(BaseModel):
    fixed_price: Optional[FixedPriceSchema] = Field(None, description="The fixed price for the subscription plan.")
    version: Optional[str] = Field(None, description="The version of the pricing scheme.")

# Billing Cycle Schema
class BillingCycleSchema(BaseModel):
    frequency: FrequencySchema = Field(..., description="The frequency of the billing cycle.")
    tenure_type: Literal['REGULAR', 'TRIAL'] = Field(..., description="The type of billing cycle tenure.")
    sequence: int = Field(..., description="The sequence of the billing cycle.")
    total_cycles: Optional[int] = Field(None, description="The total number of cycles in the billing plan.")
    pricing_scheme: PricingSchemeSchema = Field(..., description="The pricing scheme for the billing cycle.")

# Setup Fee Schema
class SetupFeeSchema(BaseModel):
    currency_code: Optional[Literal['USD']] = Field(None, description="The currency code for the setup fee.")
    value: Optional[str] = Field(None, description="The value of the setup fee.")

# Payment Preferences Schema
class PaymentPreferencesSchema(BaseModel):
    auto_bill_outstanding: Optional[bool] = Field(None, description="Indicates whether to automatically bill outstanding amounts.")
    setup_fee: Optional[SetupFeeSchema] = Field(None, description="The setup fee for the subscription plan.")
    setup_fee_failure_action: Optional[Literal['CONTINUE', 'CANCEL']] = Field(None, description="The action to take if the setup fee payment fails.")
    payment_failure_threshold: Optional[int] = Field(None, description="The number of failed payments before the subscription is canceled.")

# Taxes Schema
class TaxesSchema(BaseModel):
    percentage: Optional[str] = Field(None, description="The tax percentage.")
    inclusive: Optional[bool] = Field(None, description="Indicates whether the tax is inclusive.")

# Create Subscription Plan Parameters
class CreateSubscriptionPlanParameters(BaseModel):
    product_id: str = Field(..., description="The ID of the product for which to create the plan.")
    name: str = Field(..., description="The subscription plan name.")
    description: Optional[str] = Field(None, description="The subscription plan description.")
    billing_cycles: List[BillingCycleSchema] = Field(..., description="The billing cycles of the plan.")
    payment_preferences: PaymentPreferencesSchema = Field(..., description="The payment preferences for the subscription plan.")
    taxes: Optional[TaxesSchema] = Field(None, description="The tax details.")

# List Subscription Plans Parameters
class ListSubscriptionPlansParameters(BaseModel):
    product_id: Optional[str] = Field(None, description="The ID of the product for which to get subscription plans.")
    page: Optional[int] = Field(None, description="The page number of the result set to fetch.")
    page_size: Optional[int] = Field(None, description="The number of records to return per page (maximum 100).")
    total_required: Optional[bool] = Field(None, description="Indicates whether the response should include the total count of plans.")

# Show Subscription Plan Details Parameters
class ShowSubscriptionPlanDetailsParameters(BaseModel):
    plan_id: str = Field(..., description="The ID of the subscription plan to show.")

# Name Schema
class NameSchema(BaseModel):
    given_name: Optional[str] = Field(None, description="The subscriber given name.")
    surname: Optional[str] = Field(None, description="The subscriber last name.")

# Address Schema
class AddressSchema(BaseModel):
    address_line_1: str = Field(..., description="The first line of the address.")
    address_line_2: Optional[str] = Field(None, description="The second line of the address.")
    admin_area_1: str = Field(..., description="The city or locality.")
    admin_area_2: str = Field(..., description="The state or province.")
    postal_code: str = Field(..., description="The postal code.")
    country_code: Literal['US'] = Field(..., description="The country code.")

# Shipping Address Schema
class ShippingAddressSchema(BaseModel):
    name: Optional[NameSchema] = Field(None, description="The subscriber shipping address name.")
    address: Optional[AddressSchema] = Field(None, description="The subscriber shipping address.")

# Payment Method Schema
class PaymentMethodSchema(BaseModel):
    payer_selected: Literal['PAYPAL', 'CREDIT_CARD'] = Field(..., description="The payment method selected by the payer.")
    payee_preferred: Optional[Literal['IMMEDIATE_PAYMENT_REQUIRED', 'INSTANT_FUNDING_SOURCE']] = Field(None, description="The preferred payment method for the payee.")

# Shipping Amount Schema
class ShippingAmountSchema(BaseModel):
    currency_code: Literal['USD'] = Field(..., description="The currency code for the shipping amount.")
    value: str = Field(..., description="The value of the shipping amount.")

# Subscriber Schema
class SubscriberSchema(BaseModel):
    name: Optional[NameSchema] = Field(None, description="The subscriber name.")
    email_address: Optional[str] = Field(None, description="The subscriber email address.")
    shipping_address: Optional[ShippingAddressSchema] = Field(None, description="The subscriber shipping address.")

# Application Context Schema
class ApplicationContextSchema(BaseModel):
    brand_name: str = Field(..., description="The brand name.")
    locale: Optional[str] = Field(None, description="The locale for the subscription.")
    shipping_preference: Optional[Literal['SET_PROVIDED_ADDRESS', 'GET_FROM_FILE']] = Field(None, description="The shipping preference.")
    user_action: Optional[Literal['SUBSCRIBE_NOW', 'CONTINUE']] = Field(None, description="The user action.")
    return_url: str = Field(..., description="The return URL after the subscription is created.")
    cancel_url: str = Field(..., description="The cancel URL if the user cancels the subscription.")
    payment_method: Optional[PaymentMethodSchema] = Field(None, description="The payment method details.")

# Create Subscription Parameters
class CreateSubscriptionParameters(BaseModel):
    plan_id: str = Field(..., description="The ID of the subscription plan to create.")
    quantity: Optional[int] = Field(None, description="The quantity of the product in the subscription.")
    shipping_amount: Optional[ShippingAmountSchema] = Field(None, description="The shipping amount for the subscription.")
    subscriber: Optional[SubscriberSchema] = Field(None, description="The subscriber details.")
    application_context: Optional[ApplicationContextSchema] = Field(None, description="The application context for the subscription.")

# Show Subscription Details Parameters
class ShowSubscriptionDetailsParameters(BaseModel):
    subscription_id: str = Field(..., description="The ID of the subscription to show details.")

class Reason(BaseModel):
    reason: str = Field(..., description="Reason for Cancellation.")

# Cancel Subscription Parameters
class CancelSubscriptionParameters(BaseModel):
    subscription_id: str = Field(..., description="The ID of the subscription to cancel.")
    payload: Reason = Field(..., description="Reason for cancellation.")




CREATE_INVOICE_PROMPT = """
Create Invoices on PayPal.

This function is used to create an invoice in the PayPal system. It allows you to generate a new invoice, specifying details such as customer information, items, quantities, pricing, and tax information. Once created, an invoice can be sent to the customer for payment.
"""

LIST_INVOICE_PROMPT = """
List invoices from PayPal.

This function retrieves a list of invoices with optional pagination parameters.
"""

GET_INVOICE_PROMPT = """
Get an invoice from PayPal.

This function retrieves details of a specific invoice using its ID.
"""

SEND_INVOICE_PROMPT = """
Send an invoice to the recipient(s).

This function sends a previously created invoice to its intended recipients.
"""

SEND_INVOICE_REMINDER_PROMPT = """
Send a reminder for an invoice.

This function sends a reminder for an invoice that has already been sent but hasn't been paid yet.
"""

CANCEL_SENT_INVOICE_PROMPT = """
Cancel a sent invoice.

This function cancels an invoice that has already been sent to the recipient(s).
"""

GENERATE_INVOICE_QRCODE_PROMPT = """
Generate a QR code for an invoice.

This function generates a QR code for an invoice, which can be used to pay the invoice using a mobile device or scanning app.
"""


"""
PayPal Agent Toolkit for CrewAI.
"""

from typing import List, Optional
from pydantic import PrivateAttr, BaseModel

from ..shared.api import PayPalAPI
from ..shared.tools import tools
from ..shared.configuration import Configuration, is_tool_allowed
from .tool import PayPalTool


class PayPalToolkit:
    _tools: List = PrivateAttr()
    SOURCE = "CREWAI"
    def __init__(
        self, client_id: str, secret: str, configuration: Optional[Configuration] = None
    ):
        super().__init__()
        self._tools = []
        self.context = configuration.context if configuration and configuration.context else Configuration.Context.default()
        self.context.source = self.SOURCE
        paypal_api = PayPalAPI(client_id=client_id, secret=secret, context=self.context)

        filtered_tools = [
            tool for tool in tools if is_tool_allowed(tool, configuration)
        ]
        for tool in filtered_tools:
            args_schema = tool.get("args_schema")
    
            # Validate it's a subclass of BaseModel
            if args_schema and not issubclass(args_schema, BaseModel):
                raise TypeError(f"args_schema for '{tool['method']}' must be a Pydantic BaseModel")

            self._tools.append(
                PayPalTool(
                    name=tool["method"],
                    description=tool["description"],
                    method=tool["method"],
                    paypal_api=paypal_api,
                    args_schema=args_schema,
                )
            )

    def get_tools(self) -> List:
        """Return a list of enabled PayPal tools."""
        return self._tools




CREATE_PRODUCT_PROMPT = """
Create a product in PayPal using product catalog - create products API.
This function creates a new product that will be used in subscription plans, subscriptions.
Required Arguments are: name (product name), type (product type).
Paramters: 
    - id: (auto-generated or specify SKU of the product) The ID of the product
    - name: {product_name} (required) 
    - description: {product_description} (optional)
    - type {DIGITAL | PHYSICAL | SERVICE} (required)
    - category: {product_category} (optional) 
    - image_url: {image_url} (optional)
    - home_url: {home_url} (optional)

"""

LIST_PRODUCTS_PROMPT = """
List products from PayPal.

This function retrieves a list of products with optional pagination parameters.
"""

SHOW_PRODUCT_DETAILS_PROMPT = """
List products from PayPal.

This function retrieves a list of products with optional pagination parameters.
"""

CREATE_SUBSCRIPTION_PLAN_PROMPT = """
Create a subsctiption plan in PayPal using subscription - create plan API.
This function creates a new subscription plan that defines pricing and billing cycle details for subscriptions.
Required parameters are: product_id (the ID of the product for which to create the plan), name (subscription plan name), billing_cycles (billing cycle details).
High level: product_id, name, description, taxes, status: {CREATED|INACTIVE|ACTIVE}, billing_cycles, payment_preferences are required in json object.
While creating billing_cycles object, trial(second) billing cycle should precede regular billing cycle.
"""

LIST_SUBSCRIPTION_PLANS_PROMPT = """
List subscription plans from PayPal.

This function retrieves a list of subscription plans with optional product filtering and pagination parameters.
"""

SHOW_SUBSCRIPTION_PLAN_DETAILS_PROMPT = """
Show subscription plan details from PayPal.
This function retrieves the details of a specific subscription plan using its ID.
Required parameters are: plan_id (the ID of the subscription plan).
"""


CREATE_SUBSCRIPTION_PROMPT = """
Create a subscription in PayPal using the subscription - create subscription API.
This function allows you to create a new subscription for a specific plan, enabling the management of recurring payments.
The only required parameter is plan_id (the ID of the subscription plan). All other fields are optional and can be omitted if not provided.
The subscriber field is optional. If no subscriber information is provided, omit the subscriber field in the request payload.
The shipping address is optional. If no shipping address is provided, set the shipping_preference to GET_FROM_FILE in the application context.
The application context is also optional. If no application context information is provided, omit the application context field in the request payload.
"""

SHOW_SUBSCRIPTION_DETAILS_PROMPT = """
Show subscription details from PayPal.
This function retrieves the details of a specific subscription using its ID.
Required parameters are: subscription_id (the ID of the subscription).
"""


CANCEL_SUBSCRIPTION_PROMPT = """
Cancel a customer subscription in PayPal.

This function cancels an active subscription for a customer. It requires the subscription ID and an optional reason for cancellation.
Required parameters are: subscription_id (the ID of the subscription to be canceled).
Below is the payload request structure:
{
    "reason": "Customer requested cancellation"
}
You MUST ask the user for: 
 - subscription_id
 - reason for cancellation.

Return all of the above as structured JSON in your response.
"""


from typing import Optional, Dict, Any

class Context:

    @classmethod
    def default(cls) -> "Context":
        return cls(sandbox=True)
    
    def __init__(
        self,
        merchant_id: Optional[str] = None,
        sandbox: Optional[bool] = None,
        access_token: Optional[str] = None,
        request_id: Optional[str] = None,
        tenant_context: Optional[Any] = None,
        debug: Optional[bool] = None,
        source: Optional[str] = None,
        **kwargs: Any
    ):
        self.merchant_id = merchant_id
        self.sandbox = sandbox or False
        self.access_token = access_token
        self.request_id = request_id
        self.tenant_context = tenant_context
        self.debug = debug or False
        self.source = source or "OPEN-AI"
        self.extra = kwargs

class Configuration:
    def __init__(self, actions: Dict[str, Dict[str, bool]], context: Optional[Context] = None):
        self.actions = actions
        self.context = context

def is_tool_allowed(tool: Dict[str, Dict[str, Dict[str, bool]]], configuration: Configuration) -> bool:
    for product, product_actions in tool.get("actions", {}).items():
        for action, allowed in product_actions.items():
            if configuration.actions.get(product, {}).get(action, False):
                return True
    return False


"""PayPal Agent Toolkit."""

from typing import List, Optional
from pydantic import PrivateAttr

from ..shared.api import PayPalAPI
from ..shared.tools import tools
from ..shared.configuration import Configuration, Context, is_tool_allowed
from .tool import PayPalTool


class PayPalToolkit:
    """Toolkit for interacting with the PayPal API via tools."""

    _tools: List = PrivateAttr(default=[])
    SOURCE = "LANGCHAIN"
    def __init__(self, client_id, secret, configuration: Configuration):
        super().__init__()
        self.configuration = configuration
        self.context = configuration.context if configuration and configuration.context else Configuration.Context.default()
        self.context.source = self.SOURCE
        self._paypal_api = PayPalAPI(client_id=client_id, secret=secret, context=self.context)

        filtered_tools = [
            tool for tool in tools if is_tool_allowed(tool, configuration)
        ]

        self._tools = [
            PayPalTool(
                name=tool["method"],
                description=tool["description"],
                method=tool["method"],
                paypal_api=self._paypal_api, 
                args_schema=tool.get("args_schema"),
            )
            for tool in filtered_tools
        ]

    def get_tools(self) -> List[PayPalTool]:
        """Return a list of available PayPal tools."""
        return self._tools

    def get_paypal_api(self) -> PayPalAPI:
        """Expose the underlying PayPal API client."""
        return self._paypal_api



CREATE_ORDER_PROMPT = """
Use this tool to create a PayPal order when the user intends to purchase goods or services.

This tool is used to create a new order in PayPal. This is typically the first step in initiating a payment flow. It sets up an order with specified details such as item(s) to be purchased, quantity, amount, currency, and other details.
"""

CAPTURE_ORDER_PROMPT = """
Use this tool after the user has approved the PayPal order.

Paramters:
- order_id (str, required): The PayPal order ID provided after the user has approved the payment. 
  It typically appears in the approval link or as a token after payment approval.
"""

GET_ORDER_PROMPT = """
Use this tool to retrieve the current status of a PayPal order.

Paramters:
- order_id (str, required): The PayPal order ID you want to check.
"""




from pydantic import BaseModel, Field
from typing import List, Optional, Literal


class UnitAmount(BaseModel):
    currency_code: str = Field(..., description="Currency code of the unit amount")
    value: str = Field(..., description="The unit price. Up to 2 decimal points")


class Tax(BaseModel):
    name: Optional[str] = Field(None, description="Tax name")
    percent: Optional[str] = Field(None, description="Tax Percent")


class InvoiceItem(BaseModel):
    name: str = Field(..., description="The name of the item")
    quantity: str = Field(..., description="The quantity of the item that the invoicer provides to the payer. Value is from -1000000 to 1000000. Supports up to five decimal places. Cast to string")
    unit_amount: UnitAmount = Field(..., description="unit amount object")
    tax: Optional[Tax] = Field(None, description="tax object")
    unit_of_measure: Optional[Literal["QUANTITY", "HOURS", "AMOUNT"]] = Field(None, description="The unit of measure for the invoiced item")


class InvoicerName(BaseModel):
    given_name: Optional[str] = Field(None, description="given name of the invoicer")
    surname: Optional[str] = Field(None, description="surname of the invoicer")


class Invoicer(BaseModel):
    business_name: str = Field(..., max_length=300, description="business name of the invoicer")
    name: Optional[InvoicerName] = Field(None, description="name of the invoicer")
    email_address: Optional[str] = Field(None, description="email address of the invoicer")


class RecipientName(BaseModel):
    given_name: Optional[str] = Field(None, description="given name of the recipient")
    surname: Optional[str] = Field(None, description="surname of the recipient")


class BillingInfo(BaseModel):
    name: Optional[RecipientName] = Field(None, description="name of the recipient")
    email_address: Optional[str] = Field(None, description="email address of the recipient")


class PrimaryRecipient(BaseModel):
    billing_info: Optional[BillingInfo] = Field(None, description="The billing information of the invoice recipient")


class InvoiceDetail(BaseModel):
    invoice_date: Optional[str] = Field(None, description="The invoice date in YYYY-MM-DD format")
    currency_code: str = Field(..., description="currency code of the invoice")


class CreateInvoiceParameters(BaseModel):
    detail: InvoiceDetail = Field(..., description="The invoice detail")
    invoicer: Optional[Invoicer] = Field(None, description="The invoicer business information that appears on the invoice.")
    primary_recipients: Optional[List[PrimaryRecipient]] = Field(None, description="array of recipients")
    items: Optional[List[InvoiceItem]] = Field(None, description="Array of invoice line items")

class GetInvoiceParameters(BaseModel):
    invoice_id: str = Field(..., description="The ID of the invoice to retrieve.")


class ListInvoicesParameters(BaseModel):
    page: Optional[int] = Field(1, description="The page number of the result set to fetch.")
    page_size: Optional[int] = Field(100, ge=1, le=100, description="The number of records to return per page (maximum 100).")
    total_required: Optional[bool] = Field(None, description="Indicates whether the response should include the total count of items.")


class SendInvoiceParameters(BaseModel):
    invoice_id: str = Field(..., description="The ID of the invoice to send.")
    note: Optional[str] = Field(None, description="A note to the recipient.")
    send_to_recipient: Optional[bool] = Field(None, description="Indicates whether to send the invoice to the recipient.")
    additional_recipients: Optional[List[str]] = Field(None, description="Additional email addresses to which to send the invoice.")


class SendInvoiceReminderParameters(BaseModel):
    invoice_id: str = Field(..., description="The ID of the invoice for which to send a reminder.")
    subject: Optional[str] = Field(None, description="The subject of the reminder email.")
    note: Optional[str] = Field(None, description="A note to the recipient.")
    additional_recipients: Optional[List[str]] = Field(None, description="Additional email addresses to which to send the reminder.")


class CancelSentInvoiceParameters(BaseModel):
    invoice_id: str = Field(..., description="The ID of the invoice to cancel.")
    note: Optional[str] = Field(None, description="A cancellation note to the recipient.")
    send_to_recipient: Optional[bool] = Field(None, description="Indicates whether to send the cancellation to the recipient.")
    additional_recipients: Optional[List[str]] = Field(None, description="Additional email addresses to which to send the cancellation.")


class GenerateInvoiceQrCodeParameters(BaseModel):
    invoice_id: str = Field(..., description="The invoice id to generate QR code for")
    width: int = Field(300, description="The QR code width")
    height: int = Field(300, description="The QR code height")




"""
Tool for interacting with the PayPal API.
"""

from __future__ import annotations
import json
from typing import Any, Optional, Type
from pydantic import BaseModel
from langchain.tools import BaseTool

from ..shared.api import PayPalAPI


class PayPalTool(BaseTool):
    """
    A LangChain-compatible tool that wraps a PayPal API method call.
    
    Attributes:
        paypal_api (PayPalAPI): An instance of the PayPal API client.
        method (str): The method name to invoke on the PayPal API.
        name (str): Human-readable name of the tool.
        description (str): Description of what the tool does.
        args_schema (Optional[Type[BaseModel]]): Optional argument schema for validation.
    """

    paypal_api: PayPalAPI
    method: str
    name: str = ""
    description: str = ""
    args_schema: Optional[Type[BaseModel]] = None

    def _run(self, *args: Any, **kwargs: Any) -> str:
        """
        Executes the configured PayPal API method.

        Returns:
            str: The result from the PayPal API, or an error message.
        """
        try:
            return self.paypal_api.run(self.method, kwargs)
        except Exception as e:
            return f"Error executing PayPalTool '{self.method}': {str(e)}"

    def __repr__(self):
        return f"<PayPalTool name={self.name}, method={self.method}>"




"""PayPal Agentic Toolkit."""
from typing import List
from agents import FunctionTool
from pydantic import PrivateAttr
from ..shared.tools import tools
from ..openai.tool import PayPalTool
from ..shared.paypal_client import PayPalClient
from ..shared.configuration import Configuration, is_tool_allowed
from ..shared.api import PayPalAPI

class PayPalToolkit:

    _tools: List[FunctionTool] = PrivateAttr(default=[])
    _openai_tools = []
    _paypal_api: PayPalAPI = PrivateAttr(default=None)
    SOURCE = "OPEN-AI"

    def __init__(self, client_id, secret, configuration: Configuration):
        self.configuration = configuration
        
        self.context = configuration.context if configuration and configuration.context else Configuration.Context.default()
        self.context.source = self.SOURCE
        self._paypal_api = PayPalAPI(client_id=client_id, secret=secret, context=self.context)

        filtered_tools = [
            tool for tool in tools if is_tool_allowed(tool, configuration)
        ]

        self._tools = [
            PayPalTool(self._paypal_api, tool)
            for tool in filtered_tools
        ]
  
        self._openai_tools = [
            {
                "type": "function",
                "function": {
                    "name": tool["method"],
                    "description": tool["description"],
                    "parameters": tool["args_schema"].model_json_schema(),
                }
            }
            for tool in filtered_tools
        ]
        
    def get_openai_chat_tools(self):
        """Get the tools in the openai chat assistant."""
        return self._openai_tools
    
    def get_paypal_api(self):
        return self._paypal_api
    
    def get_tools(self) -> List[FunctionTool]:
        """Get the tools in the openai agent."""
        return self._tools



"""
This tool allows agents to interact with the PayPal API.
"""

from __future__ import annotations

from typing import Any, Optional, Type
from pydantic import BaseModel

from crewai_tools import BaseTool
from ..shared.api import PayPalAPI


class PayPalTool(BaseTool):
    """CrewAI-compatible tool for interacting with the PayPal API."""

    paypal_api: PayPalAPI
    method: str
    name: str = ""
    description: str = ""
    args_schema: Optional[Type[BaseModel]] = None

    def _run(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        """Use the PayPal API to run an operation."""
        return self.paypal_api.run(self.method, kwargs)
